#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from spec_cleaner import RpmException
from spec_cleaner.fileutils import open_datafile, open_stringio_spec


class TestFileutils(object):

    """
    We run few tests to ensure fileutils class works fine
    """

    def test_open_assertion(self):
        with pytest.raises(RpmException):
            open_stringio_spec('missing-file.txt')

    def test_open_datafile_assertion(self):
        with pytest.raises(RpmException):
            open_datafile('missing-file.txt')

    def test_open(self):
        data = open_stringio_spec('tests/fileutils-tests.py')
        data.close()

    def test_open_datafile(self):
        data = open_datafile('excludes-bracketing.txt')
        data.close()



# vim: set ts=4 sw=4 et: coding=UTF-8

"""Cleanup classes that drop most of the content."""
from typing import IO

from .rpmsection import Section


class RpmClean(Section):

    """Remove clean section."""

    def output(self, fout: IO[str], newline: bool = True, new_class_name: str = None) -> None:
        """Do not output anything here"""
        pass


class RpmChangelog(Section):

    """Remove changelog entries."""

    def add(self, line: str) -> None:
        """Only add the first line of changelog.

        This translates to adding just %changelog.
        """
        if len(self.lines) == 0:
            Section.add(self, '%changelog')



from typing import Optional

from .rpmexception import RpmException
from .rpmregexp import Regexp


class RpmRequiresToken(object):
    """
    Class containing information about the dependency token.

    Process dependencies like Requires, Recommends, Suggests, Supplements, Enhances, Conflicts. Can be used to specify
    all the values present on the line. Later on we use this to do various conversions.

    This class uses the following format and naming:

    prefix            name          operator   version
    BuildRequires:    boringpackage >=         5.2.8
    """

    comments: Optional[str] = None

    def __init__(self, name: str, operator: Optional[str] = None, version: Optional[str] = None,
                 prefix: Optional[str] = None) -> None:
        self.prefix = prefix
        self.name = name
        self.operator = operator
        self.version = version

    @staticmethod
    def _format_operator(operator: str) -> str:
        """
        Make sure the operators look sane and not use all permutations.

        Args:
            operator: A string representing the operator used in the dependency token.

        Returns:
            Formatted operator.
        """
        operator = operator.replace('=<', '<=')
        operator = operator.replace('=>', '>=')
        operator = operator.replace('==', '=')
        return operator

    @staticmethod
    def _format_name(name: str) -> str:
        """
        Make sure the name looks sane and make various replacements.

        Args:
            name: A string representing the name used in the dependency token.

        Returns:
            Formatted name.
        """
        # we just rename pkgconfig names to one unified one working everywhere
        if name == 'pkgconfig(pkg-config)' or name == 'pkg-config':
            name = 'pkgconfig'

        # omit legacy 'otherproviders' codeblock
        match = Regexp.re_otherproviders.match(name)
        if match:
            name = match.group(1)

        # replace 'packageand(pkgA:pkgB)' with '(pkgA and pkgB)' - new in RPM 4.13
        match = Regexp.re_packageand.match(name)
        if match:
            name = f'({match.group(1)} and {match.group(2)})'
        return name

    def __str__(self) -> str:
        """
        Output it all on nice pretty line.

        Returns:
            A string with a formatted output.

        Raises:
            RpmException if prefix or name is not defined or the version is defined but no operator is present.
        """

        self.name = self._format_name(self.name)
        if not self.prefix:
            raise RpmException(
                'No defined prefix in RequiresToken: prefix "{0}" name "{1}" operator "{2}" version "{3}"'.format(
                    self.prefix, self.name, self.operator, self.version
                )
            )
        if not self.name:
            raise RpmException(
                'No defined name in RequiresToken: prefix "{0}" name "{1}" operator "{2}" version "{3}"'.format(
                    self.prefix, self.name, self.operator, self.version
                )
            )
        string = self.prefix + self.name
        if (self.version and not self.operator) or (not self.version and self.operator):
            raise RpmException('Have defined version and no operator or vice versa')
        if self.version and self.operator:
            self.operator = self._format_operator(self.operator)
            string += ' ' + self.operator + ' ' + self.version

        return string



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmsection import Section


class RpmInstall(Section):

    """
    A class providing methods for %install section cleaning.

    Remove commands that wipe out the build root.
    Replace %makeinstall (suse-ism).
    """

    def add(self, line: str) -> None:
        line = self._complete_cleanup(line)

        # we do not want to cleanup buildroot, it is already clean
        if self.reg.re_clean.search(line):
            return

        line = self.reg.re_jobs.sub(' %{?_smp_mflags}', line)
        if not self.minimal:
            line = self._replace_remove_la(line)
            line = self._replace_install_command(line)

        Section.add(self, line)

    def _replace_install_command(self, line: str) -> str:
        """
        Replace various install commands with one unified mutation.

        Args:
            line: A string representing a line to process.

        Return:
            The line with install commands replaced.
        """
        make_install = '%make_install'

        # do not use install macros as we have trouble with it for now
        # we can convert it later on
        if self.reg.re_install.match(line):
            line = make_install

        # we can deal with additional params for %makeinstall so replace that
        line = line.replace('%makeinstall', make_install)

        return line

    def _replace_remove_la(self, line: str) -> str:
        """
        Replace all known variations of .la files deletion with one unified.

        Args:
            line: A string representing a line to process.

        Return:
            The processed line.
        """
        if (self.reg.re_rm.search(line) and len(self.reg.re_rm_double.split(line)) == 1) or (
            self.reg.re_find.search(line) and len(self.reg.re_find_double.split(line)) == 2
        ):
            line = 'find %{buildroot} -type f -name "*.la" -delete -print'
        return line



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from glob import glob
import os
from shutil import copyfile

import pytest
from spec_cleaner import RpmException
from spec_cleaner import RpmSpecCleaner


def collect_tests(directory):
    """
    Generate a list of tests we are going to use according to what is on hdd.

    Args:
        directory: A string representing a directory with tests.

    Return:
        A list with test names in the given directory.
    """
    testglob = os.path.join('tests', directory, '*.spec')
    return [os.path.basename(f) for f in glob(testglob)]


class TestCompare(object):
    """
    We run individual tests to verify the content compared to expected results.

    All input test files are stored in 'in' directory. Then the test outputs are divided to categories in separate
    output directories:

    E.g.:
      * 'out' - all outputs (no particular spec-cleaner option is used),
      * 'out-minimal' - outputs for tests using '--minimal' option,
      * 'web' - outputs for tests requiring internet connection,
      * 'tex', 'perl', 'cmake' - outputs for tests using '--tex', '--perl' or '--cmake' option,
      ...

    These output directories are used to help collect and run tests with the same parameters.
    """

    option_presets = {
        'pkgconfig': False,
        'inline': False,
        'diff': False,
        'diff_prog': 'vimdiff',
        'minimal': False,
        'no_curlification': False,
        'no_copyright': True,
        'libexecdir': True,
        'copyright_year': 2013,
        'remove_groups': False,
        'tex': False,
        'perl': False,
        'cmake': False,
        'keep_space': False,
    }

    @pytest.fixture(scope='function')
    def tmpfile(self, tmpdir_factory):
        tmpfile = tmpdir_factory.mktemp('run', numbered=True).join('testing.spec')
        return str(tmpfile)

    def _run_individual_test(self, test, compare_dir, infile=None, outfile=None, options=None):
        """
        Run the cleaner as specified and store the output for further comparison.
        """
        if not infile:
            infile = os.path.join('tests', 'in', test)

        full_options = {'specfile': infile, 'output': outfile}
        full_options.update(self.option_presets)
        full_options.update(options)

        cleaner = RpmSpecCleaner(full_options)
        cleaner.run()

        if compare_dir:
            compare = os.path.join('tests', compare_dir, test)
            testfile = full_options['inline'] and infile or outfile
            with open(compare) as ref, open(testfile) as test:
                assert ref.read() == test.read()

    def _compare_and_rerun(self, test, compare_dir, tmpfile, options=None):
        self._run_individual_test(test, compare_dir, None, tmpfile, options)
        tmpfile_rerun = tmpfile + '_rerun'
        self._run_individual_test(test, compare_dir, tmpfile, tmpfile_rerun, options)

    @pytest.mark.parametrize('test', collect_tests('out'))
    def test_normal_outputs(self, tmpfile, test):
        """
        Run all tests in 'out' directory without any particular spec-cleaner option.
        """
        self._compare_and_rerun(test, 'out', tmpfile, {'pkgconfig': True})

    @pytest.mark.parametrize('test', collect_tests('out-minimal'))
    def test_minimal_outputs(self, test, tmpfile):
        """
        Run tests in 'out-minimal' directory in a minimal mode.
        """
        self._compare_and_rerun(test, 'out-minimal', tmpfile, {'pkgconfig': True, 'minimal': True})

    @pytest.mark.parametrize('test', collect_tests('keep-space'))
    def test_keep_space_output(self, tmpfile, test):
        """
        Run tests in 'keep-space' directory with '--keep-space' option.
        """
        self._compare_and_rerun(test, 'keep-space', tmpfile, options={'keep_space': True})

    @pytest.mark.webtest
    @pytest.mark.parametrize('test', collect_tests('web'))
    def test_web_output(self, tmpfile, test):
        """
        Run tests in 'web' directory (these tests need an internet connection).
        """
        self._compare_and_rerun(test, 'web', tmpfile, {'pkgconfig': True})

    def test_inline_function(self, tmpfile):
        """
        Test an inline option.
        """
        test = 'bconds.spec'
        infile = os.path.join('tests', 'in', test)
        copyfile(infile, tmpfile)
        self._run_individual_test(test, None, infile=tmpfile, outfile='', options={'pkgconfig': True, 'inline': True})

    def test_diff_function(self, tmpfile):
        """
        Test passing an incorrect '--diff_prog' option.
        """
        test = 'bconds.spec'
        with pytest.raises(RpmException):
            self._run_individual_test(test, None, outfile='', options={'diff': True, 'diff_prog': 'error'})

    def test_unicode(self, tmpfile):
        """
        Test encoding.
        """
        test = 'perl-Text-Unidecode.spec'
        testpath = os.path.join('tests', 'unicode', test)
        with pytest.raises(RpmException):
            self._run_individual_test(test, None, infile=testpath, outfile='', options={'minimal': False})

    @pytest.mark.parametrize(
        'test, compare_dir, options',
        [
            ('header.spec', 'header', {'minimal': True, 'no_copyright': False}),
            ('pkgconfrequires.spec', 'out', {'pkgconfig': False}),
            ('tex.spec', 'tex', {'tex': True}),
            ('perl.spec', 'perl', {'perl': True}),
            ('cmake.spec', 'cmake', {'cmake': True}),
            ('langpackage.spec', 'group', {'remove_groups': True}),
        ],
    )
    def test_single_output(self, tmpfile, test, compare_dir, options):
        """
        Test various spec-cleaner options.
        """
        self._run_individual_test(test, compare_dir, outfile=tmpfile, options=options)



import logging
import re

from .rpmexception import NoMatchException
from .rpmrequirestoken import RpmRequiresToken

chunk_types = ('text', 'space', 'macro', 'operator', 'version')

state_types = ('start', 'name', 'operator', 'version')

re_brackets = {}
re_brackets['('] = re.compile(r'(' + r'\(' + r'|' + r'\)' + r'|' + r'\\(' + r'|' + r'\\)' + r'|' + r'[^\()]+' + r')')

re_brackets['{'] = re.compile(r'(' + r'\{' + r'|' + r'\}' + r'|' + r'\\{' + r'|' + r'\\}' + r'|' + r'[^\{}]+' + r')')

re_name = re.compile(r'[-A-Za-z0-9_~(){}@:;.+/*\[\]]+')
re_version = re.compile(r'[-A-Za-z0-9_~():.+]+')
re_spaces = re.compile(r'(\s*,\s*|\s+)')
re_macro_unbraced = re.compile('%[A-Za-z0-9_]{3,}')
re_version_operator = re.compile('(>=|<=|=>|=<|>|<|=)')

logger = logging.getLogger('DepParser')
# Switch to logging.DEBUG if needed
logger.setLevel(logging.ERROR)


def find_end_of_bracketed_macro(string, regex, opening, closing):
    macro = string[0:2]
    # eat '%{'
    string = string[2:]

    opened = 1
    while opened and string:
        try:
            bite, string = consume_chars(regex, string)
        except NoMatchException:
            raise Exception('unexpected parser error when looking for end of macro')

        if bite == opening:
            opened += 1
        elif bite == closing:
            opened -= 1
        macro += bite

    if opened:
        raise Exception('Unexpectedly met end of string when looking for end of macro')
    return macro, string


def consume_chars(regex, string):
    match = regex.match(string)
    if match:
        end = match.end()
        return string[0:end], string[end:]
    else:
        raise NoMatchException('Expected match failed (string: "%s", regex: "%s" )' % (string, regex.pattern))


def read_boolean(string):
    return find_end_of_bracketed_macro(string, re_brackets['('], '(', ')')


def matching_bracket(bracket):
    if bracket == '{':
        return '}'
    elif bracket == '(':
        return ')'
    raise Exception('Undefined bracket matching - add defintion of "%s" to ' 'matching_bracket()' % bracket)


def read_macro(string):
    opening = string[1]
    closing = matching_bracket(opening)
    return find_end_of_bracketed_macro(string, re_brackets[opening], opening, closing)


def read_next_chunk(string):
    chunk = ''
    chunk_type = ''

    if not string:
        return '', '', 'text'

    if string[0:2] in ('>=', '<=', '=>', '=<'):
        chunk = string[0:2]
        chunk_type = 'operator'
        rest = string[2:]

    elif string[0:1] in ('<', '>', '='):
        chunk = string[0:1]
        chunk_type = 'operator'
        rest = string[1:]

    elif string[0].isspace() or string[0] == ',':
        chunk = ''
        chunk_type = 'space'
        rest = consume_chars(re_spaces, string)[1]

    elif string[0:2] == '%%':
        chunk = '%%'
        chunk_type = 'text'
        rest = string[2:]

    elif string[0:2] in ('%{', '%('):
        chunk, rest = read_macro(string)
        chunk_type = 'macro'

    elif string[0] == '%':
        chunk, rest = consume_chars(re_macro_unbraced, string)
        chunk_type = 'macro'

    elif string[0] == '(':
        chunk, rest = read_boolean(string)
        chunk_type = 'macro'

    else:
        chunk, rest = consume_chars(re_name, string)
        chunk_type = 'text'

    return (rest, chunk, chunk_type)


class DepParserError(Exception):
    pass


class DependencyParser:
    def __init__(self, line):
        # adding comma will cause flush in the end of line
        self.string = line + ', '
        self.parsed = []
        self.token = []
        self.state = 'start'
        self.space = False
        self.token_name = ''
        self.token_operator = None
        self.token_version = None
        self.go_on = True
        while self.go_on:
            self.string, self.next, self.next_type = read_next_chunk(self.string)
            logger.debug(
                """========
                chunk: '%s'
                chunk_type: '%s'
                rest: '%s'
                token: '%s'
                parsed: '%s'""",
                self.next,
                self.next_type,
                self.string,
                self.token,
                self.parsed,
            )
            self.state_change()

    def flat_out(self):
        result = []
        for name, operator, ver in self.parsed:
            result.append(RpmRequiresToken(name, operator, ver))
        return result

    def flush(self):
        self.parsed.append((self.token_name, self.token_operator, self.token_version))
        # cleanup state
        self.token = []
        self.token_name = ''
        self.token_operator = None
        self.token_version = None
        if not self.string:
            self.go_on = False

    def reconstitute_token(self):
        r = ''.join(self.token)
        logger.debug("reconstituting '%s'", r)
        self.token = []
        return r

    def name_state_change(self):
        if self.next_type in ['text', 'macro']:
            if self.space:
                logger.debug('text after space --> flush')
                self.token_name = self.reconstitute_token()
                self.space = False
                self.flush()
        elif self.next_type == 'space':
            self.space = True
        elif self.next_type == 'operator':
            self.token_name = self.reconstitute_token()
            self.state = 'operator'
            self.space = False

        self.token.append(self.next)

    def operator_state_change(self):
        if self.next_type in ['text', 'macro']:
            self.state = 'version'
            self.token_operator = self.reconstitute_token()
            self.space = False
        elif self.next_type == 'space':
            self.space = True
        elif self.next_type == 'operator':
            if self.space:
                raise DepParserError('found operator after operator')

        self.token.append(self.next)

    def version_state_change(self):
        if self.next_type == 'text':
            pass
        elif self.next_type == 'space':
            self.token_version = self.reconstitute_token()
            self.flush()
            self.state = 'name'
            self.space = False
        elif self.next_type == 'macro':
            pass
        elif self.next_type == 'operator':
            raise DepParserError('found operator after version')

        self.token.append(self.next)

    def start_state_change(self):
        if self.next_type == 'text':
            self.state = 'name'
        elif self.next_type == 'space':
            pass
        elif self.next_type == 'macro':
            self.state = 'name'
        elif self.next_type == 'operator':
            raise DepParserError('found operator when name expected')

        self.token.append(self.next)

    def state_change(self):
        if self.state == 'name':
            self.name_state_change()
        elif self.state == 'operator':
            self.operator_state_change()
        elif self.state == 'version':
            self.version_state_change()
        elif self.state == 'start':
            self.start_state_change()
        logger.debug('new state: %s', self.state)



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmsection import Section


class RpmCheck(Section):

    """
    A class providing methods for %check section cleaning.

    Replace various troublemakers in check phase.
    """

    def add(self, line: str) -> None:
        line = self._complete_cleanup(line)

        # smp_mflags for jobs macro replacement
        line = self.reg.re_jobs.sub('%{?_smp_mflags}', line)

        if not self.minimal:
            line = self._add_jobs(line)
            line = self._replace_pytest(line)

        Section.add(self, line)

    def _add_jobs(self, line: str) -> str:
        """
        Add %{?_smp_mflags} to 'make' call.

        Args:
            line: A string representing a line to process.

        Return:
            The processed line.
        """
        # add jobs if we have just make call on line
        # if user want single thread he should specify -j1
        if self.reg.re_make.match(line):
            # if there are no smp_flags or jobs spec
            if line.find('%{?_smp_mflags}') == -1 and line.find('-j') == -1:
                # Don't append %_smp_mflags if the line ends with a backslash,
                # it would break the formatting
                if not line.endswith('\\') and not line.lstrip().startswith('#'):
                    line = self.reg.re_make.sub(r'\1make %{?_smp_mflags}\2', line)
        return line

    def _replace_pytest(self, line: str) -> str:
        """
        Replace various pytest calls with %pytest or %pytest_arch macros.

        Args:
            line: A string representing a line to process.

        Return:
            The processed line.
        """
        line = self.reg.re_pytest.sub('%pytest', line)
        line = self.reg.re_pytest_arch.sub('%pytest_arch', line)
        return line



# vim: set ts=4 sw=4 et: coding=UTF-8

import os.path
import shlex
import subprocess
import sys
import tempfile
from typing import Any, Dict, IO, List, Optional, Type

from .fileutils import open_stringio_spec
from .rpmbuild import RpmBuild
from .rpmcheck import RpmCheck
from .rpmcopyright import RpmCopyright
from .rpmdescription import RpmDescription
from .rpmexception import RpmException
from .rpmfiles import RpmFiles
from .rpmhelpers import (
    find_macros_with_arg,
    load_keywords_whitelist,
    parse_rpm_showrc,
    read_cmake_changes,
    read_group_changes,
    read_licenses_changes,
    read_perl_changes,
    read_pkgconfig_changes,
    read_tex_changes,
)
from .rpminstall import RpmInstall
from .rpmpackage import RpmPackage
from .rpmpreamble import RpmPreamble
from .rpmprep import RpmPrep
from .rpmprune import RpmChangelog, RpmClean
from .rpmregexp import Regexp
from .rpmscriplets import RpmScriptlets
from .rpmsection import Section


class RpmSpecCleaner(object):

    """
    Class wrapping all section parsers responsible for ensuring
    that all sections are checked and accounted for.
    If the section is required and not found it is created with
    blank values as fixme for the spec creator.

    Attributes:
        specfile: A string with the path to the specfile to process.
        fin: An in-memory input stream with the input file data.
        fout: A file object representing output file.
        current_section: A Section object representing current section of spec file.
        skip_run: A bool indicating whether the cleaning of the specfile should be skipped.
        options: A dictionary holding both spec-cleaner commandline arguments and auxiliary options.
        reg: A Regexp object that holds all regexps that will be used in spec-cleaner.
        section_starts: A list of tuples where the first item is regex object representing a start of the specfile
                        section and the second is a corresponding class that should handle it.
        _previous_line: A string holding the previous line of the currently processed specfile.
        _previous_nonempty_line: A string holding a nonempty previous line of the currently processed specfile.
    """
    specfile: Optional[str] = None
    current_section: Section
    skip_run: bool = False
    _previous_line: Optional[str] = None
    _previous_nonempty_line: Optional[str] = None

    def __init__(self, options: Dict[str, Any]) -> None:

        """Initialize and load options into the RpmSpecCleaner object and runs prep methods.

        Args:
            options: A dictionary holding spec-cleaner command line options.
        """
        self.options = options

        # Initialize main license and subpkg option
        self.options['license'] = None
        self.options['subpkglicense'] = False

        # Compile keywords for unbracing
        self.options['unbrace_keywords'] = self._unbrace_keywords()

        # Load all the remaining file operations
        self.options['tex_conversions'] = []
        self.options['pkgconfig_conversions'] = []
        self.options['cmake_conversions'] = []
        self.options['perl_conversions'] = []
        if self.options['tex']:
            self.options['tex_conversions'] = read_tex_changes()
        if self.options['pkgconfig']:
            self.options['pkgconfig_conversions'] = read_pkgconfig_changes()
        if self.options['cmake']:
            self.options['cmake_conversions'] = read_cmake_changes()
        if self.options['perl']:
            self.options['perl_conversions'] = read_perl_changes()
        self.options['license_conversions'] = read_licenses_changes()
        if self.options['remove_groups']:
            self.options['allowed_groups'] = None
        else:
            self.options['allowed_groups'] = read_group_changes()
        self.options['reg'] = Regexp(self.options['unbrace_keywords'])

        # If gvim is used for the diff then run it in foreground mode
        if self.options['diff_prog'].startswith('gvim') and ' -f' not in self.options['diff_prog']:
            self.options['diff_prog'] += ' -f'

        self.reg = self.options['reg']
        self.fin = open_stringio_spec(self.options['specfile'])

        # Section starts detection
        self.section_starts = [
            (self.reg.re_spec_package, RpmPackage),
            (self.reg.re_spec_description, RpmDescription),
            (self.reg.re_spec_prep, RpmPrep),
            (self.reg.re_spec_build, RpmBuild),
            (self.reg.re_spec_install, RpmInstall),
            (self.reg.re_spec_clean, RpmClean),
            (self.reg.re_spec_check, RpmCheck),
            (self.reg.re_spec_scriptlets, RpmScriptlets),
            (self.reg.re_spec_triggers, RpmScriptlets),
            (self.reg.re_spec_files, RpmFiles),
            (self.reg.re_spec_changelog, RpmChangelog),
        ]

        # Find all the present licenses
        self._load_licenses()

        # Determine if we need to skip the spec ('#nospeccleaner' tag)
        self._find_skip_parser()

        # Set what will be the output of the cleaning
        self._select_mode()

    def _select_mode(self) -> None:
        """
        Set up what will be the output of the cleaning process.

        Based on the options given to the commandline possible options are: output file, inline or a diff program
        showing differences.
        """
        self.fout: IO[Any]
        if self.options['output']:
            self.fout = open(self.options['output'], 'w')
        elif self.options['inline']:
            self.fout = open(self.options['specfile'], 'w')
        elif self.options['diff']:
            self.fout = tempfile.NamedTemporaryFile(mode='w+', prefix=os.path.split(self.options['specfile'])[-1] + '.', suffix='.spec')
        else:
            self.fout = sys.stdout

    def _unbrace_keywords(self) -> List[str]:
        """
        Create a list of keywords that shouldn't be in the curly brackets.

        It searches for keywords in the whitelist file, global macro functions in 'rpm --showrc' and macro functions
        in the specfile.

        Returns:
            A list of such keywords.
        """
        keywords = load_keywords_whitelist()
        global_macrofuncs = parse_rpm_showrc()
        spec_macrofuncs = find_macros_with_arg(self.options['specfile'])
        return keywords + global_macrofuncs + spec_macrofuncs

    def _find_skip_parser(self) -> None:
        """
        Search the specfile for the user defined '#nospeccleaner' tag (means that specfile shouldn't be cleaned).

        If the tag is found then skip_run member is set to True, else it's False.
        """
        for line in self.fin:
            if self.reg.re_skipcleaner.match(line):
                self.skip_run = True
                break
        self.fin.seek(0)

    def _load_licenses(self) -> None:
        """
        Detect all present licenses in the specfile and load them into 'options' member.

        If we have more than one then put license to the each subpkg.
        """
        licenses: List[str] = []
        for line in self.fin:
            if self.reg.re_license.match(line):
                line = line.rstrip('\n')
                line = line.rstrip('\r')
                line = line.rstrip()
                match = self.reg.re_license.match(line)
                value = match.groups()[-1]
                if value not in licenses:
                    licenses.append(value)
        if len(licenses) > 1:
            self.options['subpkglicense'] = True
            # put first license as placeholder if main preamble is missing one
            self.options['license'] = licenses[0]
        self.fin.seek(0)

    def _detect_preamble_section(self, line: str) -> bool:
        """
        Detect if the line starts a preamble or not.

        Args:
            line: A string representing a line to process.

        Returns:
            True if it's a preamble, False otherwise.
        """
        # This is seriously ugly but can't think of cleaner way FIXME
        if not isinstance(self.current_section, (RpmPreamble, RpmPackage)):
            if any([re.match(line) for re in [self.reg.re_bcond_with, self.reg.re_debugpkg]]):
                return True

            # We can have locally defined variables in phases
            if not isinstance(self.current_section, (RpmInstall, RpmCheck, RpmBuild, RpmPrep)) and (
                self.reg.re_define.match(line) or self.reg.re_global.match(line)
            ):
                return True
        return False

    def _detect_condition_change(self, line: str) -> bool:
        """
        Detect if the line contains a condition change (e.g. '%endif', '%else' or the end of the code block).

        Args:
            line: A string representing a line to process.

        Returns:
             True if a condition change was found, False otherwise.
        """
        if any([re.match(line) for re in [self.reg.re_endif, self.reg.re_else, self.reg.re_endcodeblock]]):
            return True
        return False

    def _detect_new_section(self, line: str) -> Optional[Type[Section]]:
        """
        Detect if the line contains a new section (and which) or not.

        Args:
            line: A string representing a line to process.

        Returns:
            A Section (or a subclass) object that was detected.
            None if we are staying in the same section or if we have a multiline value from preamble.
        """
        # Detect if we have multiline value from preamble
        # mypy: we need to ignore type check here because mypy cannot detect that we are checking the existence of
        # the 'multiline' attribute before accessing it
        if hasattr(self.current_section, 'multiline') and self.current_section.multiline:  # type: ignore
            return None

        # Detect if we match condition and that is from global space
        # Ie like in the optional packages where if is before class definition
        # For the "if" we need to detect it more smartly:
        #   check if the current line is starting new section, and if so
        #   if previous non-empty-uncommented line was starting the condition
        # we end up the condition section in preamble (if applicable) and
        # proceed to output
        if self._detect_condition_change(line) or (
            type(self.current_section) is Section and (self.reg.re_if.match(line) or self.reg.re_codeblock.match(line))
        ):
            if not hasattr(self.current_section, 'condition') or (hasattr(self.current_section, 'condition') and not self.current_section.condition):
                # If we have to break out we go ahead with small class
                # which just print the one evil line
                return Section

        # try to verify if we start some specific section
        for (regexp, newclass) in self.section_starts:
            if regexp.match(line):
                # check if we are in if conditional and act accordingly if we
                # change sections
                if hasattr(self.current_section, 'condition') and self.current_section.condition:
                    self.current_section.condition = False
                    if hasattr(self.current_section, 'end_subparagraph'):
                        # mypy: we need to ignore type check here because mypy cannot detect that we are checking the
                        # existence of the 'end_subparagraph' attribute before accessing it
                        self.current_section.end_subparagraph(True)  # type: ignore
                return newclass

        # if we still are here and we are just doing copyright
        # and we are not on commented line anymore, just jump to Preamble
        if isinstance(self.current_section, RpmCopyright):
            if not self.reg.re_comment.match(line):
                return RpmPreamble
            # if we got two empty lines then the copyright also ended
            if self._previous_line == '' and line == '':
                self.current_section.add(line)
                return RpmPreamble

        # If we actually start matching global content again we need to
        # switch back to preamble, ie %define after %description/etc.
        if self._detect_preamble_section(line):
            return RpmPreamble

        # If we are in clean section and encounter whitespace
        # we need to stop deleting
        # This avoids deleting %if before %files section
        if isinstance(self.current_section, RpmClean) and line.strip() == '':
            return Section

        # we are staying in the section
        return None

    def _check_for_newline(self, detected_class: Optional[Type[Section]], line: str) -> bool:
        """
        Check if we want newline or not after the end of section detected.

        Args:
            detected_class: A Section class (or a subclass) representing section that needs to be checked.
            line: A string representing a line to process.

        Returns: True if we want a newline, False otherwise.
        """
        # We don't want to print newlines before %else and %endif
        if detected_class == Section and self._detect_condition_change(line):
            return False
        else:
            # We also do not want to print newline if at the end of the
            # previous section we actually had a commentary, ie comment above
            # new section
            if self._previous_line and self._previous_line.startswith('#'):
                return False
            else:
                return True

    def run(self) -> None:
        """ The main spec-cleaner method.

        Raises:
            RpmException if a diff program can't be executed.
        """
        # If we are skipping the specfile we should do nothing
        if self.skip_run:
            sys.stderr.write(".spec file {0} is not being processed due to definiton of 'nospeccleaner'\n".format(self.options['specfile']))
            for line in self.fin:
                self.fout.write(line)
            self.fout.flush()
            return

        # We always start with Copyright
        self.current_section = RpmCopyright(self.options)

        # FIXME: we need to store the content locally and then reorder
        #        to maintain the specs all the same (eg somebody put
        #        filelist to the top).
        for line in self.fin:
            # Remove newlines to make it easier to parse things
            line = line.rstrip('\n')
            line = line.rstrip('\r')

            new_class = self._detect_new_section(line)
            # Following line is debug output with class info
            # USE: 'spec-cleaner file > /dev/null' to see the stderr output
            # sys.stderr.write("class: '{0}' line: '{1}'\n".format(new_class, line))
            if new_class:
                self.current_section.output(self.fout, self._check_for_newline(new_class, line), new_class.__name__)
                # start new class
                self.current_section = new_class(self.options)
                # skip empty line adding if we are switching sections
                if self._previous_line == '' and line == '':
                    continue

            # Do not store data from clean and skip out here
            if isinstance(self.current_section, RpmClean):
                continue

            self.current_section.add(line)
            self._previous_line = line
            if line != '' and not line.startswith('#'):
                self._previous_nonempty_line = line

        # no need to not output newline at the end even for minimal -> no condition
        self.current_section.output(self.fout)
        # add changelog at the end of the file
        if not isinstance(self.current_section, RpmChangelog) and self._previous_nonempty_line != '%changelog':
            self.fout.write('%changelog\n')
        self.fout.flush()

        # if the '--diff' option was used, run the diff program
        if self.options['diff']:
            cmd = shlex.split(
                self.options['diff_prog'] + ' ' + self.options['specfile'].replace(' ', '\\ ') + ' ' + self.fout.name.replace(' ', '\\ ')
            )
            try:
                subprocess.call(cmd, shell=False)
            except OSError as error:
                raise RpmException('Could not execute %s (%s)' % (self.options['diff_prog'].split()[0], error.strerror))

    def __del__(self) -> None:
        """
        Close the input and output files.
        """

        if self.fin:
            self.fin.close()
        if self.fout:
            self.fout.close()



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmsection import Section


class RpmFiles(Section):

    """
    A class providing methods for %files section cleaning.
    """
    def add(self, line: str) -> None:
        line = self._complete_cleanup(line)
        line = self.strip_useless_spaces(line)
        line = self._remove_doc_on_man(line)
        line = self._move_license_from_doc(line)
        # we only get empty %doc left over
        if line == '%doc ':
            return

        if not self.minimal:
            # prune obsolete defattr that is default
            if self.reg.re_defattr.match(line):
                return
            line = self._set_man_compression(line)

        # toss out empty lines if there are more than one in succession
        if line == '' and (not self.previous_line or self.previous_line == ''):
            return

        Section.add(self, line)

    @staticmethod
    def _remove_doc_on_man(line: str) -> str:
        """
        Remove all "%doc %_mandir" to -> "%_mandir" as it is pointless to do twice.

        Args:
            line: A string representing a line to process.

        Returns:
            The processed line.
        """
        line = line.replace('%doc %{_mandir}', '%{_mandir}', 1)
        line = line.replace('%doc %{_infodir}', '%{_infodir}', 1)
        return line

    def _set_man_compression(self, line: str) -> str:
        """
        Set proper compression suffix on man/info pages, instead of .gz/.* use
        the proper macro variable.

        Args:
           line: A string representing a line to process.

        Returns:
            The processed line.
        """
        if line.startswith('%{_mandir}'):
            line = self.reg.re_man_compression.sub(r'\1%{?ext_man}', line)
        if line.startswith('%{_infodir}'):
            line = self.reg.re_info_compression.sub('.info%{?ext_info}', line)
        return line

    def _move_license_from_doc(self, line: str) -> str:
        """
        Move license file from %doc to %license.

        Args:
           line: A string representing a line to process.

        Returns:
            The processed line.
        """
        if line.startswith('%doc') and self.reg.re_doclicense.search(line):
            licences = ''
            match = self.reg.re_doclicense.search(line)
            while match:
                licences += match.group()
                line = self.reg.re_doclicense.sub('', line, 1)
                match = self.reg.re_doclicense.search(line)
            Section.add(self, '%license {}'.format(licences))
        return line



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmsection import Section


class RpmPrep(Section):

    """
    A class providing methods for %prep section cleaning.

    It simplifies %setup and %patch lines.
    """

    def add(self, line):
        line = self._complete_cleanup(line)
        line = self._cleanup_setup(line)
        if not self.minimal:
            line = self._prepare_patch(line)
        Section.add(self, line)

    def _cleanup_setup(self, line: str) -> str:
        """
        Remove the useless stuff from %setup line.

        Args:
            line: A string representing a line to process.

        Return:
            The processed line.
        """
        # NOTE: not using regexp as this covers 99% cases for now
        if line.startswith('%setup'):
            line = line.replace(' -qn', ' -q -n')
            line = line.replace(' -q', '')
            line = self.reg.re_setup.sub(' ', line)
            line = self.strip_useless_spaces(line)
            line = line.replace('%setup', '%setup -q')

        return line

    def _prepare_patch(self, line: str) -> str:
        """
        Convert patchlines to something pretty.

        E.g. it converts "%patch -P 50 -p10" to "%patch50 -p10" and so on.

        Args:
            line: A string representing a line to process.

        Return:
            The line with pretty patchlines.
        """
        # -p0 is default
        line = line.replace('-p0', '')
        # %patch0 is desired
        if (line.startswith('%patch ') or line == '%patch') and '-P' not in line:
            line = line.replace('%patch', '%patch0')

        # convert the %patch -P 50 -p10 to %patch50 -p10
        # this apply only if there is ONE -P on the line, not multiple ones
        match = self.reg.re_patch_prep.match(line)
        if match:
            line = self.strip_useless_spaces('%%patch%s %s %s' % (match.group(2), match.group(1), match.group(3)))

        return line



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmexception import RpmException
from .rpmhelpers import add_group, find_pkgconfig_declaration, find_pkgconfig_statement, fix_license, sort_uniq
from .rpmrequirestoken import RpmRequiresToken


class RpmPreambleElements(object):
    """
    Class containing structure used in rpmpreamble.
    List of all the elements possible to be provided in dict and list forms.
    """

    category_to_key = {
        'name': 'Name',
        'version': 'Version',
        'release': 'Release',
        'license': 'License',
        'summary': 'Summary',
        # The localized summary can contain various values, so it can't be here
        'url': 'URL',
        'group': 'Group',
        'source': 'Source',
        'nosource': 'NoSource',
        'patch': 'Patch',
        'buildrequires': 'BuildRequires',
        'buildconflicts': 'BuildConflicts',
        'buildignores': '#!BuildIgnore',
        'conflicts': 'Conflicts',
        'prereq': 'PreReq',
        'requires': 'Requires',
        'requires_eq': '%requires_eq',
        'requires_ge': '%requires_ge',
        'recommends': 'Recommends',
        'suggests': 'Suggests',
        'enhances': 'Enhances',
        'supplements': 'Supplements',
        # Provides/Obsoletes cannot be part of this since we want to keep them
        # mixed, so we'll have to specify the key when needed
        'buildarch': 'BuildArch',
        'exclusivearch': 'ExclusiveArch',
        'excludearch': 'ExcludeArch',
    }

    categories_order = (
        'define',
        'bconds',
        'bcond_conditions',
        'name',
        'version',
        'release',
        'summary',
        'summary_localized',
        'license',
        'group',
        'url',
        'source',
        'nosource',
        'patch',
        'patternprovides',  # fake provides with special function for patterns
        'patternobsoletes',  # fake provide/obsoletes with old pattern symbols
        'patterncodeblock',  # fake condition placement for patterns to stay on top
        'buildrequires',
        'buildconflicts',
        'buildignores',
        'requires',
        'requires_eq',
        'requires_ge',
        'prereq',
        'requires_phase',  # this is Requires(pre/post/...)
        'recommends',
        'suggests',
        'enhances',
        'supplements',
        'conflicts',
        'provides_obsoletes',
        'buildarch',
        'exclusivearch',
        'excludearch',
        'misc',
        'build_conditions',
        'conditions',
        'tail',
    )

    # categories that are sorted based on value in them
    categories_with_sorted_package_tokens = [
        'patternprovides',
        'buildrequires',
        'buildconflicts',
        'buildignores',
        'prereq',
        'requires',
        'requires_eq',
        'requires_ge',
        'requires_phase',
        'recommends',
        'suggests',
        'enhances',
        'supplements',
        'conflicts',
    ]

    # categories that are sorted based on key value (eg Patch0 before Patch1)
    categories_with_sorted_keyword_tokens = ('source', 'patch')

    def __init__(self, options):
        self.items = {}
        for i in self.categories_order:
            self.items[i] = []
        self.current_group = []
        # minimal mode
        self.minimal = options['minimal']
        # regexp object
        self.reg = options['reg']
        # pkgconfig requirement detection
        self.br_pkgconfig_required = False
        # license string
        self.license = options['license']
        # dict of license replacement options
        self.license_conversions = options['license_conversions']
        # initialize list of groups that need to pass over conversion fixer
        self.categories_with_package_tokens = self.categories_with_sorted_package_tokens[:]
        # these packages actually need fixing after we sent the values to
        # reorder them
        self.categories_with_package_tokens.append('provides_obsoletes')

    def _sort_helper_key(self, a):
        if isinstance(a, str) or isinstance(a, RpmRequiresToken):
            key = str(a)
        elif isinstance(a, list):
            # if this is a list then all items except last are comment or whitespace
            key = str(a[-1])
        else:
            raise RpmException('Unknown type during sort: %s' % a)

        # Special case is the category grouping where we have to get the number in
        # after the value
        if self.reg.re_patch.match(key):
            match = self.reg.re_patch.match(key)
            key = int(match.group(2))
        elif self.reg.re_source.match(key):
            match = self.reg.re_source.match(key)
            value = match.group(1)
            if not value:
                value = '1'
            key = int(value)
        # Put brackety ()-style deps at the end of the list, after all other
        elif self.reg.re_brackety_requires.search(key):
            key = '2' + key
        elif self.reg.re_patternmacro.search(key):
            key = '0' + key
        else:
            key = '1' + key
        return key

    def _insert_value(self, category, value, key=None):
        """
        Add value to specified keystore
        """
        key = self.compile_category_prefix(category, key)
        line = RpmRequiresToken(value, None, None, key)
        self.items[category].append(line)

    def _add_pkgconfig_buildrequires(self, nested):
        """
        Check the content of buildrequires and add pkgconfig as an item
        in case there are any pkgconfig() style dependencies present

        If we are in the top level object for preamble we append the BR,
        otherwise we do just verify if there are nay dependencies
        """
        # first generate flat list from the BR
        buildrequires = []
        for group in self.items['buildrequires']:
            buildrequires += add_group(group)
        # Check if we need the pkgconfig
        if not self.br_pkgconfig_required and find_pkgconfig_statement(buildrequires):
            self.br_pkgconfig_required = True
        # only in case we are in main scope
        if not nested:
            if self.br_pkgconfig_required and not find_pkgconfig_declaration(buildrequires):
                self._insert_value('buildrequires', 'pkgconfig')

    @staticmethod
    def _verify_prereq_message(elements):
        """
        Verify if the prereq is present in the Requires(*) and add the fixme
        comment if needed
        """
        message = '# FIXME: use proper Requires(pre/post/preun/...)'

        prereq_found = False
        message_found = False

        # Check first if we have prereq values included
        for element in elements:
            if isinstance(element, RpmRequiresToken):
                if element.prefix.startswith('PreReq'):
                    prereq_found = True
                    break
        if not prereq_found:
            return elements

        # Verify the message is not already present
        for element in elements:
            if isinstance(element, str):
                if element.startswith(message):
                    message_found = True
        if message_found:
            return elements

        # add the message on the first position after any whitespace
        location = next(i for i, j in enumerate(elements) if j)
        elements.insert(location, message)

        return elements

    @staticmethod
    def _remove_duplicates(elements):
        """
        Remove duplicate requires/buildrequires/etc
        """
        results = []
        for element in elements:
            match = False
            # anything else than requirestoken
            if not isinstance(element, RpmRequiresToken):
                results.append(element)
                continue
            # no results stored yet
            if not results:
                results.append(element)
                continue
            # search already stored content
            for index, item in enumerate(results):
                # if item is string we didn't match
                if not isinstance(item, RpmRequiresToken):
                    continue
                # names and prefix must always match
                if item.name == element.name and item.prefix == element.prefix:
                    # do we have full match on everything
                    if item.version == element.version and item.operator == element.operator:
                        # append comment if needed only as we are 100% match
                        if element.comments:
                            tmp = results[index]
                            if tmp.comments:
                                tmp.comments += element.comments
                            else:
                                tmp.comments = element.comments
                            results[index] = tmp
                        match = True
                        break
                    # new one specifies version
                    if not item.version and element.version:
                        if item.comments:
                            if element.comments:
                                element.comments += item.comments
                            else:
                                element.comments = item.comments
                        results[index] = element
                        match = True
                        break
                    # for version determination which could be ommited one
                    # must use rpm versionCompare to get same results
                    # unfortunately it uses too many resources so we simply
                    # leave this to the maintainer
            if not match:
                results.append(element)
        return results

    def _run_global_list_operations(self, phase, elements):
        """
        Run all the checks that need to be run on the finalized sorted list
        rather than on invidiual value
        """
        # check if we need to add comment for the prereq
        if not self.minimal and phase == 'prereq':
            elements = self._verify_prereq_message(elements)

        return elements

    def compile_category_prefix(self, category, key=None):
        """
        Simply compile the category key and provide enough whitespace for the
        values to be alligned
        """
        keylen = len('BuildRequires:  ')

        if category == 'tail':
            return ''
        elif key:
            pass
        elif category in self.category_to_key:
            key = self.category_to_key[category]
        else:
            raise RpmException('Unhandled category in preamble: %s' % category)

        # append : only if the thing is not known macro
        if not key.startswith('%'):
            key += ':'
        # if the key is already longer then just add one space
        if len(key) >= keylen:
            key += ' '
        # fillup rest of the alignment if key is shorter than muster
        while len(key) < keylen:
            key += ' '
        return key

    def flatten_output(self, needs_license=False, nested=False):
        """
        Do the finalized output for the itemlist.
        """
        lines = []
        elements = []

        # add license to the package if missing and needed
        if needs_license and not self.items['license']:
            self.license = fix_license(self.license, self.license_conversions)
            self._insert_value('license', self.license)
        # add pkgconfig dep
        self._add_pkgconfig_buildrequires(nested)
        # remove duplicates
        for i in self.categories_with_package_tokens:
            self.items[i] = self._remove_duplicates(self.items[i])
        for i in self.categories_order:
            sorted_list = []
            if i in self.categories_with_sorted_package_tokens:
                self.items[i].sort(key=self._sort_helper_key)
            # sort-out within the ordered groups based on the key
            if i in self.categories_with_sorted_keyword_tokens:
                self.items[i].sort(key=self._sort_helper_key)
                self.items[i] = sort_uniq(self.items[i])
            # flatten the list from list of lists as no reordering is planned
            for group in self.items[i]:
                sorted_list += add_group(group)
            # now do all sorts of operations where we needed sorted lists
            lines += self._run_global_list_operations(i, sorted_list)
        if self.current_group:
            # the current group was not added to any category. It's just some
            # random stuff that should be at the end anyway.
            lines += add_group(self.current_group)
            self.current_group = []

        for line in lines:
            elements.append(str(line))
        return elements



# vim: set ts=4 sw=4 et: coding=UTF-8

# Copyright (c) 2015, SUSE LINUX Products GmbH, Nuernberg, Germany
# All rights reserved.
# See COPYING for details.

import argparse
from datetime import datetime
import os
import sys
from typing import Any, Dict, List

from .rpmcleaner import RpmSpecCleaner
from .rpmexception import RpmException, RpmWrongArgs


__version__ = '1.1.6'


def process_args(argv: List[str]) -> Dict[str, Any]:
    """
    Parse and process commandline arguments.

    Args:
        argv: A list of passed arguments.

    Returns:
        A dict mapping arguments to the corresponding values.

    Raises:
        RpmWrongArgs: If the specfile doesn't exist or
                      if the output file already exists but '--force' option (overwrite the output) wasn't used.
    """
    parser = argparse.ArgumentParser(
        prog='spec-cleaner',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Cleans the given spec file according to style guide and returns the result. If "#nospeccleaner"'
                    'tag is used in the spec file than the spec file will not be cleaned.',
    )

    # Make the -d, -i, and -o exclusive as we can do only one of those
    output_group = parser.add_mutually_exclusive_group()

    parser.add_argument('specfile', metavar='SPEC', type=str, help='spec file to beautify')
    parser.add_argument(
        '-c',
        '--cmake',
        action='store_true',
        help='convert dependencies to their cmake() counterparts, requires bit more of cleanup in spec afterwards.',
    )
    output_group.add_argument('-d', '--diff', action='store_true', help='run the diff program to show differences between new and original specfile.')
    parser.add_argument('--diff-prog', default='vimdiff', help='specify the diff binary to call with diff option.')
    parser.add_argument('-f', '--force', action='store_true', help='overwrite the output file if already exist.')
    output_group.add_argument('-i', '--inline', action='store_true', help='inline the changes directly to the parsed file.')
    parser.add_argument(
        '-m', '--minimal', action='store_true', help='run in minimal mode that does not do anything intrusive (ie. just sets the Copyright)'
    )
    parser.add_argument(
        '--no-curlification', action='store_true', help='do not convert variables bracketing (%%{macro}) and keep it as it was on the input'
    )
    parser.add_argument('--no-copyright', action='store_true', help='do not include official SUSE copyright hear and just keep what is present')
    parser.add_argument('--libexecdir', action='store_true', help='convert /usr/lib to %%{_libexecdir}')
    parser.add_argument('--remove-groups', action='store_true', help='remove groups from the specfile.')
    parser.add_argument(
        '--copyright-year',
        metavar='YYYY',
        type=int,
        default=datetime.now().year,
        help='year to insert into the copyright header when re-generating it',
    )
    output_group.add_argument('-o', '--output', default='', help='specify the output file for the cleaned spec content.')
    parser.add_argument(
        '-p',
        '--pkgconfig',
        action='store_true',
        help='convert dependencies to their pkgconfig() counterparts, requires bit more of cleanup in spec afterwards.',
    )
    parser.add_argument('--perl', action='store_true', help='convert dependencies to their perl() counterparts, highly expansive use carefully')
    parser.add_argument(
        '-t', '--tex', action='store_true', help='convert dependencies to their tex() counterparts, requires verification of the output afterwards.'
    )
    parser.add_argument('-v', '--version', action='version', version=__version__, help='show package version and exit')
    parser.add_argument('-k', '--keep-space', action='store_true', help='keep empty lines in preamble intact.')

    # print help if there is no argument
    if len(argv) < 1:
        parser.print_help()
        sys.exit(0)

    options = parser.parse_args(args=argv)

    # the spec must exist for us to do anything
    if not os.path.exists(options.specfile):
        raise RpmWrongArgs('{0} does not exist.'.format(options.specfile))

    # the path for output must exist and the file must not be there unless
    # force is specified
    if options.output:
        options.output = os.path.expanduser(options.output)
        if not options.force and os.path.exists(options.output):
            raise RpmWrongArgs('{0} already exists.'.format(options.output))

    # convert options to dict
    options_dict = vars(options)
    return options_dict


def main() -> int:
    """
    Main function that calls argument parsing ensures their sanity
    and then creates RpmSpecCleaner object that works with passed spec file.
    """

    try:
        options = process_args(sys.argv[1:])
    except RpmWrongArgs as exception:
        sys.stderr.write('ERROR: {0}\n'.format(exception))
        return 1

    try:
        cleaner = RpmSpecCleaner(options)
        cleaner.run()
    except RpmException as exception:
        sys.stderr.write('ERROR: {0}\n'.format(exception))
        return 1

    return 0



# vim: set ts=4 sw=4 et: coding=UTF-8
from typing import Any, Dict, IO, List, Optional

from .rpmregexp import Regexp


class Section(object):

    """
    Basic object for parsing each section of spec file.

    It stores the lines in a list and remembers content of
    previous line at hand.
    The unbrace_keywords content is passed from creating
    object to reduce calculation price.
    Various functions do replacement of common typos to
    unify all the content.

    Attributes:
        lines: A list of all lines of the section.
        previous_line: A string that represents the previous line.
        spec: A string with the path to the processed specfile.
        minimal: A flag indicating whether we run in minimal mode (no intrusive operations).
        no_curlification: A flag indicating whether we want to convert variables to curly brackets.
        libexecdir: A flag indicating whether we want convert /usr/lib to %%{_libexecdir}.
        reg: A Regexp object that holds all regexps that will be used in spec-cleaner.
        condition: A flag representing if we are in the conditional or not.
        _condition_counter: An int for counting in how many (nested) condition we currently are.
    """

    def __init__(self, options: Dict[str, Any]) -> None:
        self.lines: List[str] = []
        self.previous_line: Optional[str] = None
        self.spec: str = options['specfile']
        self.minimal: bool = options['minimal']
        self.no_curlification: bool = options['no_curlification']
        self.libexecdir: bool = options['libexecdir']
        self.reg: Regexp = options['reg']
        # Are we inside of conditional or not
        self.condition: bool = False
        self._condition_counter: int = 0

    def _complete_cleanup(self, line: str) -> str:
        """
        Call all the cleanups in proper order so it
        can be called beforehand if we override add phase and want to
        do our replaces after cleanup.

        Args:
            line: A string representing a line to process.

        Returns:
            The cleaned line.
        """
        line = line.rstrip()
        # remove nbsp for normal spaces
        line = line.replace(u'\xa0', ' ')

        if not line.startswith('#'):
            if not self.minimal and not self.no_curlification:
                line = self.embrace_macros(line)
            line = self.replace_buildroot(line)
            line = self.replace_optflags(line)
            line = self.replace_known_dirs(line)
            line = self.replace_utils(line)
            line = self.replace_buildservice(line)
            line = self.replace_preamble_macros(line)
            line = self.replace_python_expand(line)

        return line

    def _check_conditions(self, line: str) -> None:
        """
        Set 'condition' member to True if we are in condition that is contained (False otherwise).

        Also adjusts '_condition_counter' member according to the fact whether we enter or leave the condition.

        Args:
            line: A string representing a line to process.
        """
        if self.reg.re_if.match(line) or self.reg.re_codeblock.match(line):
            self._condition_counter += 1
        if self.reg.re_endif.match(line) or self.reg.re_endcodeblock.match(line):
            self._condition_counter -= 1

        if self._condition_counter > 0:
            self.condition = True
        else:
            self.condition = False

    def add(self, line: str) -> None:
        """
        Run the cleanup of the line and add the line to the list of lines.

        Args:
            line: A string representing a line to process.
        """
        line = self._complete_cleanup(line)

        # conditions detect
        self._check_conditions(line)

        # append to the file
        self.lines.append(line)
        self.previous_line = line

    def output(self, fout: IO[str], newline: bool = True, new_class_name: str = None) -> None:
        """
        Manage printing of the section.

        Always append one empty line at the end if it is not present and a changelog is a trailing part of our spec
        so do not put nothing bellow. Also if we are jumping away just after writing one macroed line.

        Args:
            fout: A file object representing the output file.
            newline: A flag indicating whether we want to add a newline.
            new_class_name: A string with the Section name.
        """
        # we don't want to create new line
        if newline and len(self.lines) >= 1:
            if (
                self.lines[-1] != ''
                and self.lines[-1] != '%changelog'
                and not self.lines[-1].startswith('%if')
                and not self.lines[-1].startswith('%pre')
                and not self.lines[-1].startswith('%post')
                and not self.lines[-1].endswith('\\')
            ):
                self.lines.append('')
            if new_class_name != 'RpmScriptlets' and (self.lines[-1].startswith('%pre') or self.lines[-1].startswith('%post')):
                self.lines.append('')
            # remove the newlines around ifs if they are not wanted
            if len(self.lines) >= 2:
                if self.lines[-1] == '' and (self.lines[-2].startswith('%if') or self.lines[-2].startswith('%else')):
                    self.lines.pop()

        for line in self.lines:
            fout.write(line + '\n')

    @staticmethod
    def strip_useless_spaces(line: str) -> str:
        """
        Remove useless multiple spaces in some areas.

        It can't be called everywhere so we have to call it in
        children classes where fit.

        Args:
            line: A string representing a line to process.

        Returns:
            The line without useless spaces.
        """
        return ' '.join(line.split())

    def embrace_macros(self, line: str) -> str:
        """
        Add {} around known macros that have no arguments and are not
        on whitelist.

        Whitelist is passed from caller object.

        Args:
            line: A string representing a line to process.

        Returns:
            The line with curlified macros.
        """
        # I don't think that this can be done within one regexp replacement
        # if you have idea, send me a patch :)

        # work only with non-commented part
        sp = line.split('#')
        # so, for now, put braces around everything, what looks like macro,
        previous = sp[0]

        while True:
            sp[0] = self.reg.re_macro.sub(r'\1%{\3}\5', sp[0])
            if sp[0] == previous:
                break
            else:
                previous = sp[0]

        # and replace back known keywords to braceless state again
        sp[0] = self.reg.re_unbrace_keywords.sub(r'%\1', sp[0])
        # re-create the line back
        return '#'.join(sp)

    def replace_buildroot(self, line: str) -> str:
        """
        Replace RPM_BUILD_ROOT for buildroot.

        Args:
            line: A string representing a line to process.

        Returns:
            The processed line.
        """
        line = self.reg.re_rpmbuildroot.sub(r'%{buildroot}\2', line)
        line = self.reg.re_rpmbuildroot_quotes.sub(r'%{buildroot}', line)
        return line

    def replace_optflags(self, line: str) -> str:
        """
        Replace RPM_OPT_FLAGS for %{optflags}.

        Args:
            line: A string representing a line to process.

        Returns:
            The processed line.
        """
        # if the optflags is the only thing then also add quotes around it
        line = self.reg.re_optflags_quotes.sub('="%{optflags}"', line)
        line = self.reg.re_optflags.sub('%{optflags}', line)
        return line

    def replace_python_expand(self, line: str) -> str:
        """
        Replace classic python macros with those starting with '$' if it's used with %python_expand macro.

        E.g. with %python_expand one must use "%python_expand %{$python_sitelib}" instead of
        "%python_expand %{python_sitelib}". Known variables: python_sitearch, python_sitelib, python_version,
        python_bin_suffix and python.

        Args:
            line: A string representing a line to process.

        Returns:
            The line with python macros replaced.
        """
        if line.startswith('%python_expand') or line.startswith('%{python_expand'):
            line = self.reg.re_python_expand.sub(r'%{$\1}', line)
            line = self.reg.re_python_interp_expand.sub(r' $\1 ', line)
        return line

    def replace_known_dirs(self, line: str) -> str:
        """
        Replace hardcoded stuff like /usr/share -> %{_datadir}.

        Args:
            line: A string representing a line to process.

        Returns:
            The processed line.
        """
        line = self.reg.re_oldprefix.sub(r'%{_prefix}\1', line)
        line = self.reg.re_prefix.sub(r'%{_prefix}\1', line)
        line = self.reg.re_bindir.sub(r'%{_bindir}\1', line)
        line = self.reg.re_sbindir.sub(r'%{_sbindir}\1', line)
        if not self.minimal and self.libexecdir:
            line = self.reg.re_libexecdir.sub(r'%{_libexecdir}\1', line)
        line = self.reg.re_includedir.sub(r'%{_includedir}\1', line)
        line = self.reg.re_datadir.sub(r'%{_datadir}\1', line)
        line = self.reg.re_mandir.sub(r'%{_mandir}\1', line)
        line = self.reg.re_infodir.sub(r'%{_infodir}\1', line)
        line = self.reg.re_docdir.sub(r'%{_docdir}\1', line)
        line = self.reg.re_initdir.sub(r'%{_initddir}\1', line)
        line = self.reg.re_sysconfdir.sub(r'%{_sysconfdir}\1', line)
        line = self.reg.re_localstatedir.sub(r'%{_localstatedir}\1', line)
        line = self.reg.re_libdir.sub(r'%{_libdir}\2', line)
        line = self.reg.re_initddir.sub(r'%{_initddir}\1', line)

        return line

    def replace_utils(self, line: str) -> str:
        """
        Remove the macro calls for utilities and rather use direct commands (OBS ensures there is only one anyway).

        Args:
            line: A string representing a line to process.

        Returns:
            The line without macros for utilities.
        """
        r = {
            'id_u': 'id -u',
            'ln_s': 'ln -s',
            'lzma': 'xz --format-lzma',
            'mkdir_p': 'mkdir -p',
            'awk': 'gawk',
            'cc': 'gcc',
            'cpp': 'gcc -E',
            'cxx': 'g++',
            'remsh': 'rsh',
        }
        for i in r:
            line = line.replace('%{__' + i + '}', r[i])
            if self.minimal:
                line = line.replace('%__' + i, r[i])

        for i in (
            'aclocal',
            'ar',
            'as',
            'autoconf',
            'autoheader',
            'automake',
            'bzip2',
            'cat',
            'chgrp',
            'chmod',
            'chown',
            'cp',
            'cpio',
            'file',
            'gpg',
            'grep',
            'gzip',
            'id',
            'install',
            'ld',
            'libtoolize',
            'make',
            'mkdir',
            'mv',
            'nm',
            'objcopy',
            'objdump',
            'patch',
            'perl',
            'python',
            'python2',
            'python3',
            'pypy3',
            'ranlib',
            'restorecon',
            'rm',
            'rsh',
            'sed',
            'semodule',
            'ssh',
            'strip',
            'tar',
            'unzip',
            'xz',
        ):
            line = line.replace('%{__' + i + '}', i)
            if self.minimal:
                line = line.replace('%__' + i, i)

        line = self.reg.re_deprecated_egrep_regex.sub(r'grep -E', line)
        line = self.reg.re_deprecated_fgrep_regex.sub(r'grep -F', line)

        return line

    @staticmethod
    def replace_buildservice(line: str) -> str:
        """
        Pretty format the conditions for distribution/version detection.

        Replace %{suse_version} for 0%{?suse_version} and the like.

        Args:
            line: A string representing a line to process.

        Returns:
            The line with formatted version conditions.
        """
        for i in ['centos', 'debian', 'fedora', 'mandriva', 'meego', 'rhel', 'sles', 'suse', 'ubuntu']:
            line = line.replace('%{' + i + '_version}', '0%{?' + i + '_version}').replace('00%{', '0%{')
        return line

    def replace_preamble_macros(self, line: str) -> str:
        """
        Replace %{S:0} and %{P:0} for %{SOURCE0} and %{PATCH0}.

        Args:
            line: A string representing a line to process.

        Returns: The processed line.
        """
        line = self.reg.re_ptch.sub(r'%{PATCH\1}', line)
        line = self.reg.re_src.sub(r'%{SOURCE\1}', line)
        return line



# vim: set ts=4 sw=4 et: coding=UTF-8

from io import StringIO
import os
import sys
import sysconfig
from typing import IO

from .rpmexception import RpmException


def open_datafile(name: str) -> IO[str]:
    """
    Open data files.

    Used all around so kept glob here for importing.

    Args:
        name: A string representing the name of the datafile to open.

    Raises:
        RpmException if the file is not found in predefined datadirs.
    """
    homedir = os.getenv('HOME', '~') + '/.local/'

    possible_paths = (
        '{0}/../data/{1}'.format(os.path.dirname(os.path.realpath(__file__)), name),
        '{0}/share/spec-cleaner/{1}'.format(homedir, name),
        '{0}/share/spec-cleaner/{1}'.format(sysconfig.get_path('data'), name),
        '{0}/share/spec-cleaner/{1}'.format(sys.prefix, name),
    )

    for path in possible_paths:
        try:
            _file = open(path, mode='r')
        except OSError:
            pass
        else:
            return _file
    # file not found
    raise RpmException("File '{}' not found in datadirs".format(name))


def open_stringio_spec(name: str) -> IO[str]:
    """
    Open regular files with exception handling.

    Args:
        name: A string with the file name.

    Returns:
        A file object.

    Raises:
        RpmException if the file is not readable.
    """
    data = StringIO()
    try:
        with open(name, mode='r') as f:
            data.write(f.read())
            data.seek(0, 0)
    except (IOError, UnicodeDecodeError) as error:
        raise RpmException(str(error))
    return data



# vim: set ts=4 sw=4 et: coding=UTF-8

import os
from typing import IO

from .rpmsection import Section


class RpmCopyright(Section):

    """
    A class providing methods for copyright section cleaning.

    It always creates default SUSE copyright.
    Keeps around Copyrights of other uses and some of the build defines
    that are still relevant. Everything else is ignored.
    """

    def __init__(self, options):
        Section.__init__(self, options)
        self.no_copyright = options['no_copyright']
        self.year = options['copyright_year']
        self.copyrights = []
        self.buildrules = []
        self.my_copyright = ''
        self.vimmodeline = ''

    def _add_pkg_header(self):
        specname = os.path.splitext(os.path.basename(self.spec))[0]
        self.lines.append(
            """#
# spec file for package {0}
#""".format(
                specname
            )
        )

    def _create_default_copyright(self):
        self.my_copyright = '# Copyright (c) {0} SUSE LLC'.format(self.year)

    def _add_copyright(self):
        self._create_default_copyright()
        self.lines.append(self.my_copyright)

        for i in self.copyrights:
            self.lines.append(i)

    def _add_default_license(self):
        self.lines.append(
            """#
# All modifications and additions to the file contributed by third parties
# remain the property of their copyright owners, unless otherwise agreed
# upon. The license for this file, and modifications and additions to the
# file, is the same license as for the pristine package itself (unless the
# license for the pristine package is not an Open Source License, in which
# case the license is the MIT License). An "Open Source License" is a
# license that conforms to the Open Source Definition (Version 1.9)
# published by the Open Source Initiative.

# Please submit bugfixes or comments via https://bugs.opensuse.org/
#"""
        )

    def _add_buildrules(self) -> None:
        for i in sorted(self.buildrules):
            self.lines.append(i)

    def _add_modelines(self) -> None:
        # add vim modeline if found
        if self.vimmodeline:
            self.lines.append(self.vimmodeline)

    def add(self, line: str) -> None:
        # if we have no copyright header we actually should not touch it not
        # wipe out, thus just add everything to known lines
        if self.no_copyright:
            self.lines.append(line)
            return
        if not self.lines and not line:
            return
        copyright_match = self.reg.re_copyright_string.match(line)
        if copyright_match and not self.reg.re_suse_copyright.search(line):
            # always replace whitespace garbage on copyright line
            line = '# Copyright (c) {0}'.format(copyright_match.group(1))
            self.copyrights.append(line)
        elif self.reg.re_rootforbuild.match(line):
            self.buildrules.append('# needsrootforbuild')
        elif self.reg.re_binariesforbuild.match(line):
            self.buildrules.append('# needsbinariesforbuild')
        elif self.reg.re_nodebuginfo.match(line):
            self.buildrules.append('# nodebuginfo')
        elif self.reg.re_icecream.match(line):
            self.buildrules.append('# icecream')
        elif self.reg.re_vimmodeline.match(line):
            self.vimmodeline = line
        elif self.reg.re_sslcerts.match(line):
            self.buildrules.append('# needssslcertforbuild')
        else:
            # anything not in our rules gets tossed out
            return

    def output(self, fout: IO[str], newline: bool = True, new_class_name: str = None):
        if not self.no_copyright:
            self._add_modelines()
            self._add_pkg_header()
            self._add_copyright()
            self._add_default_license()
            self._add_buildrules()
            self.lines.append('')
            self.lines.append('')
        else:
            newline = False
        Section.output(self, fout, newline, new_class_name)



# vim: set ts=4 sw=4 et: coding=UTF-8

# We basically extend rpmcheck
from .rpmcheck import RpmCheck


class RpmBuild(RpmCheck):

    """
    A class providing methods for %build section cleaning.
    """

    def add(self, line: str) -> None:
        # we do not want to run suseupdateconfig, deprecated
        if self.reg.re_suseupdateconfig.search(line):
            return

        # if user uses cmake/configure directly just recommend him using the macros
        if not self.minimal:
            self._comment_macro_calls(line)

        RpmCheck.add(self, line)

    def _comment_macro_calls(self, line: str) -> None:
        """
        Add a comment if the package uses direct call of certain build tools instead of macro.

        Args:
            line: A string representing a line to process.
        """
        if self.reg.re_configure.match(line):
            msg = '# FIXME: you should use the %%configure macro'
        elif self.reg.re_cmake.match(line):
            msg = '# FIXME: you should use %%cmake macros'
        elif self.reg.re_meson.match(line):
            msg = '# FIXME: you should use %%meson macros'
        elif self.reg.re_qmake5.match(line):
            msg = '# FIXME: you should use %qmake5 macro'
        else:
            return

        position = len(self.lines)
        if self.previous_line and self.previous_line.endswith('\\'):
            # Backtrack until we are on top of multiline command
            for i in reversed(self.lines):
                if i.endswith('\\'):
                    position -= 1
                else:
                    # we stop backtracking on the first
                    # non-backslashed line
                    break
        # Check if on the position above is no comment
        # if not add our message
        if not self.lines[position - 1].startswith('#'):
            self.lines.insert(position, msg)



# vim: set ts=4 sw=4 et: coding=UTF-8
from typing import IO

from .rpmsection import Section


class RpmScriptlets(Section):

    """
    A class providing methods for scriptlet section cleaning.

    Do %post -p /sbin/ldconfig when only scriplet command is /sbin/ldconfig.
    """

    def add(self, line: str) -> None:
        line = self._complete_cleanup(line)
        line = self._remove_deprecated_ldconfig(line)
        Section.add(self, line)

    def _remove_deprecated_ldconfig(self, line: str) -> str:
        """
        Replace %run_ldconfig with /sbin/ldconfig.

        Args:
            line: A string representing a line to process.

        Returns:
            The processed line.
        """
        line = self.reg.re_ldconfig.sub('/sbin/ldconfig', line)
        return line

    def output(self, fout: IO[str], newline: bool = True, new_class_name: str = None) -> None:
        if not self.minimal:
            self._collapse_multiline_ldconfig()
        Section.output(self, fout, newline, new_class_name)

    def _collapse_multiline_ldconfig(self) -> None:
        """
        Merge two lines ldconfig call to the one line and adjust lines member accordingly.
        """
        # if we have 2 lines or rest of them are empty, pop those
        for i in reversed(self.lines):
            if i:
                break
            else:
                self.lines.pop()
        if len(self.lines) == 2:
            if self.lines[1] == '/sbin/ldconfig':
                pkg = self.lines[0]
                self.lines = []
                self.lines.append('{0} -p /sbin/ldconfig'.format(pkg))



# vim: set ts=4 sw=4 et: coding=UTF-8

import re
from subprocess import check_output
from typing import Dict, List

from .fileutils import open_datafile, open_stringio_spec
from .rpmexception import RpmException
from .rpmrequirestoken import RpmRequiresToken

LICENSES_CHANGES = 'licenses_changes.txt'
TEX_CONVERSIONS = 'tex_conversions.txt'
PKGCONFIG_CONVERSIONS = 'pkgconfig_conversions.txt'
PERL_CONVERSIONS = 'perl_conversions.txt'
CMAKE_CONVERSIONS = 'cmake_conversions.txt'
GROUPS_LIST = 'allowed_groups.txt'
BRACKETING_EXCLUDES = 'excludes-bracketing.txt'


def parse_rpm_showrc() -> List[str]:
    """
    Create a list of all macro functions in the 'rpm --showrc' output.

    Returns:
        A list of such macro functions.
    """
    macros: List[str] = []

    re_rc_macrofunc = re.compile(r'^-[0-9]+[:=]\s(\w+)\(.*')
    output = check_output(['rpm', '--showrc'])
    for line in output.decode().split('\n'):
        found_macro = re_rc_macrofunc.sub(r'\1', line)
        if found_macro != line:
            macros += [found_macro]
    return macros


def load_keywords_whitelist() -> List[str]:
    """
    Create a list of keywords contained in BRACKETING_EXCLUDES file (keywords that shouldn't be in brackets).

    Returns:
        A list of such keywords.
    """
    with open_datafile(BRACKETING_EXCLUDES) as f:
        return [line.rstrip('\n') for line in f]


def find_macros_with_arg(spec: str) -> List[str]:
    """
    Create a list of all macro functions in the spec file.

    Args:
        spec: A string with the path to the specfile.

    Returns:
        A list of such macro functions.
    """
    macrofuncs: List[str] = []

    re_spec_macrofunc = re.compile(r'^\s*%define\s(\w+)\(.*')
    with open_stringio_spec(spec) as f:
        for line in (l.rstrip('\n') for l in f):
            found_macro = re_spec_macrofunc.sub(r'\1', line)
            if found_macro != line:
                macrofuncs += [found_macro]
    return macrofuncs


def read_conversion_changes(conversion_file):
    with open_datafile(conversion_file) as f:
        # the values are split by  ': '
        return {key: value for key, value in (line.split(': ') for line in f)}


def read_tex_changes():
    return read_conversion_changes(TEX_CONVERSIONS)


def read_pkgconfig_changes():
    return read_conversion_changes(PKGCONFIG_CONVERSIONS)


def read_perl_changes():
    return read_conversion_changes(PERL_CONVERSIONS)


def read_cmake_changes():
    return read_conversion_changes(CMAKE_CONVERSIONS)


def read_licenses_changes() -> Dict[str, str]:
    """
    Create mapping of old licences to new licences.

    It uses LICENCES_CHANGES file that has the following format:

    correct license string<tab>known bad license string

    Tab is used as a separator.

    Returns:
        A dict with the mapping.

    """
    with open_datafile(LICENSES_CHANGES) as f:
        next(f)  # strip newline
        return {old: correct for correct, old in (line.rstrip('\n').split('\t') for line in f)}


def read_group_changes():
    with open_datafile(GROUPS_LIST) as f:
        next(f)  # header starts with link where we find the groups
        return [line.rstrip('\n') for line in f]


def fix_license(value, conversions):
    # license ; should be replaced by ands so find it
    re_license_semicolon = re.compile(r'\s*;\s*')
    value = value.rstrip(';')
    value = re_license_semicolon.sub(' and ', value)
    # split using 'or', 'and' and parenthesis, ignore empty strings
    licenses = []
    for a in re.split(r'(\(|\)| and | AND | OR | or (?!later)|;)', value):
        if a != '':
            licenses.append(a)
    if not licenses:
        licenses.append(value)

    for (index, my_license) in enumerate(licenses):
        my_license = ' '.join(my_license.split())
        my_license = my_license.replace('ORlater', 'or later')
        my_license = my_license.replace('ORsim', 'or similar')
        if my_license in conversions:
            my_license = conversions[my_license]
        licenses[index] = my_license

    # create back new string with replaced licenses
    s = ' '.join(licenses).replace('( ', '(').replace(' )', ')').replace(' and ', ' AND ').replace(' or ', ' OR ').replace(' with ', ' WITH ')
    return s


def sort_uniq(seq):
    def _check_list(x):
        if isinstance(x, list):
            return True
        else:
            return False

    seen = {}
    result = []
    for item in seq:
        marker = item
        # We can have list there with comment
        # So if list found just grab latest in the sublist
        if _check_list(marker):
            marker = marker[-1]
        if marker in seen:
            # Not a list, no comment to preserve
            if not _check_list(item):
                continue
            # Here we need to preserve comment content
            # As the list is already sorted we can count on it to be
            # seen in previous run.
            # match the current and then based on wether the previous
            # value is a list we append or convert to list entirely
            prev = result[-1]
            if _check_list(prev):
                # Remove last line of the appending
                # list which is the actual dupe value
                item.pop()
                # Remove it from orginal
                prev.pop()
                # join together
                prev += item
                # append the value back
                prev.append(marker)
                result[-1] = prev
            else:
                # Easy as there was no list
                # just replace it with our value
                result[-1] = item
            continue
        seen[marker] = 1
        result.append(item)
    return result


def add_group(group):
    """
    Flatten the lines of the group from sublits to one simple list
    """
    if isinstance(group, str):
        return [group]
    elif isinstance(group, RpmRequiresToken):
        items = []
        if group.comments:
            items += group.comments
        items.append(group)
        return items
    elif isinstance(group, list):
        items = []
        for subgroup in group:
            items += add_group(subgroup)
        return items
    else:
        raise RpmException('Unknown type of group in preamble: %s' % type(group))


def find_pkgconfig_statement(elements):
    """
    Find pkgconfig() statement.

    Args:
        elements: A list of items we want to scan.

    Returns:
        True if pkgconfig() statement was found (and pkgconfig declaration wasn't), False otherwise.
    """

    pkgconfig_found = find_pkgconfig_declaration(elements)
    for i in elements:
        if isinstance(i, RpmRequiresToken):
            if 'pkgconfig(' in i.name and not pkgconfig_found:
                return True
    return False


def find_pkgconfig_declaration(elements):
    """
    Find if there is direct pkgconfig dependency in the paragraph.

    Args:
        elements: A list of items we want to scan.

    Returns:
        True if a pkgconfig dependency was found, False otherwise.
    """
    for i in elements:
        if isinstance(i, RpmRequiresToken):
            if 'pkgconfig ' in i.name or i.name.endswith('pkgconfig'):
                return True
    return False



from glob import glob

from setuptools import setup
from spec_cleaner import __version__

setup(
    name='spec_cleaner',
    description='RPM .spec files cleaner',
    long_description=('Command-line tool for cleaning various formatting errors in RPM .spec files'),
    url='https://github.com/rpm-software-management/spec-cleaner',
    download_url='https://github.com/rpm-software-management/spec-cleaner',
    version=__version__,
    author='Tomáš Chvátal',
    author_email='tchvatal@suse.cz',
    maintainer='Tomáš Chvátal',
    maintainer_email='tchvatal@suse.cz',
    license='License :: OSI Approved :: BSD License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        'Topic :: Text Processing',
    ],
    platforms=['Linux'],
    keywords=['SUSE', 'RPM', '.spec', 'cleaner'],
    tests_require=['pytest', 'pytest-cov', 'pytest-xdist'],
    packages=['spec_cleaner'],
    data_files=[('lib/obs/service/', glob('obs/*')), ('share/spec-cleaner', glob('data/*'))],
    entry_points={'console_scripts': ['spec-cleaner = spec_cleaner:main']},
)



# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmpreamble import RpmPreamble
from .rpmsection import Section


class RpmPackage(RpmPreamble):

    """
    A class providing methods for %package section cleaning.

    We handle subpackage case as the normal preamble.
    """

    def add(self, line: str) -> None:
        # The first line (%package) should always be added and is different
        # from the lines we handle in RpmPreamble.
        if not self.previous_line:
            Section.add(self, line)
            return

        # If the package is lang package we add here comment about the lang
        # package
        if (
            len(self.lines) == 1
            and (self.previous_line.startswith('%') and (self.previous_line.endswith(' lang') or self.previous_line.endswith('-lang')))
            and not line.startswith('#')
        ):
            if not self.minimal:
                Section.add(self, '# FIXME: consider using %%lang_package macro')

        RpmPreamble.add(self, line)



# vim: set ts=4 sw=4 et: coding=UTF-8

import re
from typing import List


class Regexp(object):

    """
        Singleton containing all regular expressions compiled in one run.
        So we can use them later everywhere without compiling them again,
    """

    # section macros
    re_spec_package = re.compile(r'^%package(\s+|$)', re.IGNORECASE)
    re_spec_description = re.compile(r'^%description(\s+|$)', re.IGNORECASE)
    re_spec_prep = re.compile(r'^%prep\s*$', re.IGNORECASE)
    re_spec_build = re.compile(r'^%build\s*$', re.IGNORECASE)
    re_spec_install = re.compile(r'^%install\s*$', re.IGNORECASE)
    re_spec_clean = re.compile(r'^%clean\s*$', re.IGNORECASE)
    re_spec_check = re.compile(r'^%check\s*$', re.IGNORECASE)
    re_spec_scriptlets = re.compile(
        r'(?:^%pretrans(\s+|$))|(?:^%pre(\s+|$))|(?:^%post(\s+|$))|(?:^%verifyscript(\s+|$))|(?:^%preun(\s+|$))|(?:^%postun(\s+|$))|(?:^%posttrans(\s+|$))',
        re.IGNORECASE,
    )
    re_spec_triggers = re.compile(
        r'(?:^%filetriggerin(\s+|$))|(?:^%filetriggerun(\s+|$))|(?:^%filetriggerpostun(\s+|$))|(?:^%transfiletriggerin(\s+|$))|(?:^%transfiletriggerun(\s+|$))|(?:^%transfiletriggerpostun(\s+|$))',
        re.IGNORECASE,
    )
    re_spec_files = re.compile(r'^%files(\s+|$)', re.IGNORECASE)
    re_spec_changelog = re.compile(r'^%changelog\s*$', re.IGNORECASE)

    # rpmpreamble
    # WARNING: keep in sync with rpmcleaner Section change detection
    re_if = re.compile(r'^\s*(?:%{?if\s|%{?ifarch\s|%{?ifnarch\s|%{?if\S*}?(\s.*|)$)', re.IGNORECASE)
    re_codeblock = re.compile(r'^\s*((### COMMON-([a-zA-Z0-9]+)-BEGIN ###|# MANUAL BEGIN|# SECTION)(\s.*|)|# MANUAL)$', re.IGNORECASE)
    re_else = re.compile(r'^\s*%else(\s.*|)$', re.IGNORECASE)
    re_endif = re.compile(r'^\s*%endif(\s.*|)$', re.IGNORECASE)
    re_endcodeblock = re.compile(r'^\s*(### COMMON-([a-zA-Z0-9]+)-END ###|# MANUAL END|# /MANUAL|# (END|/)SECTION)(\s.*|)$', re.IGNORECASE)
    re_name = re.compile(r'^Name:\s*(\S*)', re.IGNORECASE)
    re_version = re.compile(r'^Version:\s*(.*)', re.IGNORECASE)
    re_release = re.compile(r'^Release:\s*(\S*)', re.IGNORECASE)
    re_license = re.compile(r'^License:\s*(.*)', re.IGNORECASE)
    re_summary = re.compile(r'^Summary:\s*(.*)', re.IGNORECASE)
    re_summary_localized = re.compile(r'^Summary(\(\S+\)):\s*(.*)', re.IGNORECASE)
    re_url = re.compile(r'^Url:\s*(\S*)', re.IGNORECASE)
    re_group = re.compile(r'^Group:\s*(.*)', re.IGNORECASE)
    re_vendor = re.compile(r'^Vendor:\s*(.*)', re.IGNORECASE)
    re_source = re.compile(r'^Source(\d*):\s*(.*)', re.IGNORECASE)
    re_nosource = re.compile(r'^NoSource:\s*(.*)', re.IGNORECASE)
    re_patch = re.compile(r'^((?:#[#\s]*)?)Patch(\d*):\s*(\S*)', re.IGNORECASE)
    re_buildrequires = re.compile(r'^(BuildRequires|BuildPreReq):\s*(.*)', re.IGNORECASE)
    re_buildconflicts = re.compile(r'^BuildConflicts:\s*(.*)', re.IGNORECASE)
    re_buildignores = re.compile(r'^#!BuildIgnore:\s*(.*)', re.IGNORECASE)
    re_prereq = re.compile(r'^PreReq:\s*(.*)', re.IGNORECASE)
    re_requires = re.compile(r'^Requires:\s*(.*)', re.IGNORECASE)
    re_requires_phase = re.compile(r'^Requires(\([^)]+\)):\s*(.*)', re.IGNORECASE)
    re_recommends = re.compile(r'^Recommends:\s*(.*)', re.IGNORECASE)
    re_suggests = re.compile(r'^Suggests:\s*(.*)', re.IGNORECASE)
    re_enhances = re.compile(r'^Enhances:\s*(.*)', re.IGNORECASE)
    re_supplements = re.compile(r'^Supplements:\s*(.*)', re.IGNORECASE)
    re_conflicts = re.compile(r'^Conflicts:\s*(.*)', re.IGNORECASE)
    re_provides = re.compile(r'^Provides:\s*(.*)', re.IGNORECASE)
    re_obsoletes = re.compile(r'^Obsoletes:\s*(.*)', re.IGNORECASE)
    re_buildroot = re.compile(r'^\s*BuildRoot:', re.IGNORECASE)
    re_buildarch = re.compile(r'^\s*BuildArch(itectures)?:\s*(.*)', re.IGNORECASE)
    re_exclusivearch = re.compile(r'^\s*ExclusiveArch(itectures)?:\s*(.*)', re.IGNORECASE)
    re_excludearch = re.compile(r'^\s*ExcludeArch(itectures)?:\s*(.*)', re.IGNORECASE)
    re_epoch = re.compile(r'^\s*Epoch:\s*(.*)', re.IGNORECASE)
    re_icon = re.compile(r'^\s*Icon:\s*(.*)', re.IGNORECASE)
    re_copyright = re.compile(r'^\s*Copyright:\s*(.*)', re.IGNORECASE)
    re_packager = re.compile(r'^\s*Packager:\s*(.*)', re.IGNORECASE)
    re_define = re.compile(r'^\s*%define\s*(.*)', re.IGNORECASE)
    re_global = re.compile(r'^\s*%global\s*(.*)', re.IGNORECASE)
    re_bcond_with = re.compile(r'^\s*%bcond_with(out)?\s*(.*)', re.IGNORECASE)
    re_autoreqprov = re.compile(r'^\s*AutoReqProv:.*$', re.IGNORECASE)
    re_debugpkg = re.compile(r'^%{?(debug_package|___debug_install_post)}?\s*$', re.IGNORECASE)
    re_py_requires = re.compile(r'^%{?\??py_requires}?\s*$', re.IGNORECASE)
    re_mingw = re.compile(r'^\s*%{?_mingw.*$', re.IGNORECASE)
    re_patterndefine = re.compile(r'^\s*%{?pattern_\S+}?\s*$', re.IGNORECASE)
    re_patternmacro = re.compile(r'pattern(-\S+)?\(\)', re.IGNORECASE)
    re_patternobsolete = re.compile(r'patterns-openSUSE-\S+', re.IGNORECASE)
    re_tail_macros = re.compile(r'^%{?python_subpackages}?')
    re_preamble_prefix = re.compile(r'^Prefix:\s*(.*)', re.IGNORECASE)
    # grab all macros with rpm call that query for version, this still might
    # be bit too greedy but it is good enough now
    re_rpm_command = re.compile(r'%\(\s*(rpm|echo\s+`rpm).*--queryformat\s+\'%{?VERSION}?\'.*\)')
    re_requires_eq = re.compile(r'^\s*(%{\?requires_eq:\s*)?%requires_eq\s*(.*)')
    re_requires_ge = re.compile(r'^\s*(%{\?requires_ge:\s*)?%requires_ge\s*(.*)')
    re_onelinecond = re.compile(r'^\s*%{!?[^?]*\?[^:]+:[^}]+}')
    # Special bracketed deps dection
    re_brackety_requires = re.compile(r'(pkgconfig|cmake|perl|tex|rubygem)\(')
    re_version_separator = re.compile(r'(\S+)((\s*[<>=\s]+)(\S+))*')
    # packageand(pkg1:pkg2)
    re_packageand = re.compile(r'^packageand\(\s*(\S+)\s*:\s*(\S+)\s*\)\s*$')
    # otherproviders(foo)
    re_otherproviders = re.compile(r'^otherproviders\(\s*(\S+)\s*\)\s*$')

    # rpmdescription
    re_authors = re.compile(r'^\s*Author(s)?:\s*')

    # rpmbuild
    re_jobs = re.compile(r'%{?(_smp_mflags|\?_smp_flags|\?jobs:\s*-j\s*%(jobs|{jobs}))}?')
    re_make = re.compile(r'(^\s*)make(\s.*|)$')
    re_optflags_quotes = re.compile(r'=\s*\${?RPM_OPT_FLAGS}?\s*$')
    re_optflags = re.compile(r'\${?RPM_OPT_FLAGS}?')
    re_suseupdateconfig = re.compile(r'%{?\??suse_update_config')
    re_configure = re.compile(r'(^|(.*\s)?)./configure(\s.*|)$')
    re_cmake = re.compile(r'(^|(.*\s)?)cmake(\s.*|)$')
    re_qmake5 = re.compile(r'(^|(.*\s)?)qmake-qt5(\s.*|)$')
    re_meson = re.compile(r'(^|(.*\s)?)meson(\s.*|)$')
    re_pytest = re.compile(r'%python_(expand|exec)\s+(PYTHONPATH=%{buildroot}%{\$?python_sitelib}\s+)?(\$?python\s+)?(%{_bindir}/?|-m\s+)?py\.?test(-(%{\$?python_version}|%{\$?python_bin_suffix})?)?(\s+(-v|-o addopts=-v))?')
    re_pytest_arch = re.compile(r'%python_(expand|exec)\s+(PYTHONPATH=%{buildroot}%{\$?python_sitearch}\s+)?(\$?python\s+)?(%{_bindir}/?|-m\s+)?py\.?test(-(%{\$?python_version}|%{\$?python_bin_suffix})?)?(\s+(-v|-o addopts=-v))?')
    re_python_expand = re.compile(r'%{?(python_sitelib|python_sitearch|python_bin_suffix|python_version)}?')
    re_python_interp_expand = re.compile(r'\s+(python)\s+')

    # rpmcopyright
    re_copyright_string = re.compile(r'^#\s*Copyright\ \(c\)\s*(.*)', re.IGNORECASE)
    re_suse_copyright = re.compile(r'SUSE LINUX (Products )?GmbH, Nuernberg, Germany.\s*$', re.IGNORECASE)
    re_rootforbuild = re.compile(r'^#\s*needsrootforbuild\s*$', re.IGNORECASE)
    re_binariesforbuild = re.compile(r'^#\s*needsbinariesforbuild\s*$', re.IGNORECASE)
    re_nodebuginfo = re.compile(r'^#\s*nodebuginfo\s*$', re.IGNORECASE)
    re_sslcerts = re.compile(r'^#\s*needssslcertforbuild\s*$', re.IGNORECASE)
    re_icecream = re.compile(r'^#\s*icecream\s*$', re.IGNORECASE)
    re_vimmodeline = re.compile(r'^#\s*vim:', re.IGNORECASE)
    re_skipcleaner = re.compile(r'^#\s*nospeccleaner\s*$', re.IGNORECASE)

    # rpminstall
    re_clean = re.compile(r'rm\s+(-?\w?\ ?)*"?(%{buildroot}|\$b)"?$')
    re_install = re.compile(r'{0}*(%{{makeinstall}}|make{0}+install){0}*$'.format(r'(DESTDIR=%{buildroot}|%{\?_smp_mflags}|\s|V=1|VERBOSE=1|-j\d+)'))
    re_rm = re.compile(r'rm\s+(-?\w?\ ?)*"?(%{buildroot}|\$b)"?/?"?%{_lib(dir)?}.*\*\.la;?$')
    re_find = re.compile(r'find\s+"?(%{buildroot}|\$b)("?\S?/?)*\s*.*\s+-i?name\s+["\'\\]?\*\.la($|.*[^\\]$)')
    re_find_double = re.compile(r'-i?name')
    re_rm_double = re.compile(r'(\.|{)a')

    # rpmprep
    re_patch_prep = re.compile(r'^%patch\s*([^P]*)-P\s*(\d*)\s*([^P]*)$')
    re_setup = re.compile(r'\s*-n\s+"?%{name}-%{version}"?($|\s)')

    # rpmfiles
    re_man_compression = re.compile(r'(\d)(\.?\*|\.gz|%{?ext_man}?)$')
    re_info_compression = re.compile(r'\.info(\.?\*|\.gz|%{?ext_info}?)$')
    re_defattr = re.compile(r'^\s*%defattr\s*\(\s*-\s*,\s*root\s*,\s*root\s*(,\s*-\s*)?\)\s*')
    re_doclicense = re.compile(r'(\S+)?(LICEN(S|C)E|COPYING)(\*|\.(\*|\S+))?($|\s)', re.IGNORECASE)

    # rpmscriptlets
    re_ldconfig = re.compile(r'(^|(.*\s)?)%{?run_ldconfig}?(\s.*|)$', re.IGNORECASE)
    # patches/sources
    re_ptch = re.compile(r'%{P:(\d+)}')
    re_src = re.compile(r'%{S:(\d+)}')

    # comment detection
    re_comment = re.compile(r'^$|^\s*#')

    # macro detection
    re_macro = re.compile(
        # find start of macro:
        #   either beggining of string or something which is not '%' or :
        #   where : is used after macro declaration we should not curlify
        r'(^|([^%:]))'
        +
        # macro itself:
        # '%' followed by either number not starting with '0'
        # or by chars where first is a-z or A-Z or underscore
        r'%([1-9]\d*|[a-zA-Z_]\w*'
        +
        # possibly followed by parens
        r'(\s*\([^)]*\))?'
        +
        # beyond the end of the macro
        r')(|(\W))'
    )

    # cleaning path regexps
    endmacro = r'([/\s%"]|$)'
    re_oldprefix = re.compile(r'%{?_exec_prefix}?' + endmacro)
    re_prefix = re.compile(r'(?<!\w)/usr' + endmacro)
    re_bindir = re.compile(r'%{?_prefix}?/bin' + endmacro)
    re_sbindir = re.compile(r'%{?_prefix}?/sbin' + endmacro)
    re_libexecdir = re.compile(r'%{?_prefix}?/lib' + endmacro)
    re_includedir = re.compile(r'%{?_prefix}?/include' + endmacro)
    re_datadir = re.compile(r'%{?_prefix}?/share' + endmacro)
    re_mandir = re.compile(r'%{?_datadir}?/man' + endmacro)
    re_infodir = re.compile(r'%{?_datadir}?/info' + endmacro)
    re_docdir = re.compile(r'%{?_datadir}?/doc/packages' + endmacro)
    re_initdir = re.compile(r'/etc/init.d' + endmacro)
    re_sysconfdir = re.compile(r'/etc' + endmacro)
    re_localstatedir = re.compile(r'/var' + endmacro)
    re_libdir = re.compile(r'%{?_prefix}?/(%{?_lib}?|lib64)' + endmacro)
    re_initddir = re.compile(r'%{?_initrddir}?' + endmacro)
    re_rpmbuildroot = re.compile(r'(\${?RPM_BUILD_ROOT}?|"%{?buildroot}?")([/\s%]|$)')
    re_rpmbuildroot_quotes = re.compile(r'"\${?RPM_BUILD_ROOT}?"')
    # deprecated greps
    re_deprecated_egrep_regex = re.compile(r'\begrep\b')
    re_deprecated_fgrep_regex = re.compile(r'\bfgrep\b')

    def __init__(self, keywords: List[str]) -> None:
        self.re_unbrace_keywords = re.compile('%{(' + '|'.join(keywords) + ')}')
